level = {
	width = 12,
	height = 8,
	temporaryTiles = { },
	load = function(filename, number)
		level.irises = 0
		local levelFile = io.open(filename)
		while number > 1 do
			levelFile:read(level.width * level.height)
			number = number - 1
		end
		level.rows = { }
		for y = 1, level.height do
			level.rows[y] = levelFile:read(level.width)
			for x = 1, level.width do
				local tileType = string.byte(level.rows[y], x) + 1
				if tileType == tiles.IRIS or tileType == tiles.IRIS_OPEN then
					level.irises = level.irises + 1
				end
			end
		end
		levelFile:close()
	end,
	getLevelCount = function(filename)
		return 16
	end,
	-- Gets all of the times taken for particular level file.
	getTimesTaken = function(filename)
		local matchString = filename .. ":"
		local timesTaken = { }
		for line in io.lines("settings/times.ini") do
			if string.sub(line, 1, #matchString) == matchString then
				local levelNumber, timeTaken = string.sub(line, #matchString + 1):match("^%s*(%d+)%s*=%s*(%d+)%s*$")
				if levelNumber and timeTaken then
					timesTaken[levelNumber * 1] = timeTaken * 1
				end
			end
		end
		return timesTaken
	end,
	-- Sets the time taken for a particular level.
	setTimeTaken = function(filename, level, timeTaken)
		
		local newSettingsFile = { }
		local matchString = filename .. ":"
		
		-- Presume we've made a change but have not updated an existing record.
		local settingsFileChanged = true
		local settingsFileUpdated = false
		
		for line in io.lines("settings/times.ini") do
			if string.sub(line, 1, #matchString) == matchString then
				local oldLevel, oldTime = string.sub(line, #matchString + 1):match("^%s*(%d+)%s*=%s*(%d+)%s*$")
				
				-- Coerce to numbers.
				if oldLevel then oldLevel = oldLevel * 1 end
				if oldTime then oldTime = oldTime * 1 end
				
				-- Have we beaten an existing time?
				if oldLevel and oldLevel == level and oldTime then
					if  timeTaken < oldTime then
						-- We're updating an existing record.
						newSettingsFile[#newSettingsFile + 1] = filename .. ":" .. level .. "=" .. timeTaken
						settingsFileUpdated = true
					else
						-- We've previously completed this level in a better time.
						settingsFileChanged = false
					end
				else
					-- Just append the row.
					newSettingsFile[#newSettingsFile + 1] = line
				end
			else
				-- Just append the row.
				newSettingsFile[#newSettingsFile + 1] = line
			end
		end
		
		-- Do we need to append a new row?
		if settingsFileChanged and not settingsFileUpdated then
			newSettingsFile[#newSettingsFile + 1] = filename .. ":" .. level .. "=" .. timeTaken
		end
		
		-- Has anything changed?
		if settingsFileChanged then
			local settingsFile = io.open("settings/times.ini", "w+")
			for i, line in ipairs(newSettingsFile) do
				settingsFile:write(line .. "\n")
			end
			settingsFile:close()
		end
		
	end,
	getTile = function(x, y)
		return string.byte(level.rows[y + 1], x + 1) + 1
	end,
	setTile = function(x, y, tile)
		level.rows[y + 1] = string.sub(level.rows[y + 1], 1, x) .. string.char(tile - 1) .. string.sub(level.rows[y + 1], x + 2, level.width)
		tiles.images[tile]:draw(x * tiles.size, y * tiles.size)
	end,
	setTemporaryTile = function(x, y, tile)
		table.insert(level.temporaryTiles, { x = x, y = y, tile = level.getTile(x, y) })
		level.setTile(x, y, tile)
	end,
	clearTemporaryTiles = function()
		for i, temporaryTile in ipairs(level.temporaryTiles) do
			level.setTile(temporaryTile.x, temporaryTile.y, temporaryTile.tile)
		end
		level.temporaryTiles = { }
	end,
	incrementTile = function(x, y, amount)
		level.setTile(x, y, level.getTile(x, y) + amount)
	end,
	moveTile = function(x, y, dx, dy, force)
		local newX = x + dx
		local newY = y + dy
		if newX < 0 or newX >= level.width or newY < 0 or newY >= level.height then
			if force then
				level.setTile(x, y, tiles.BLANK)
				return nil
			else
				return tiles.STEEL
			end
		else
			local hitTile = level.getTile(newX, newY)
			if force or hitTile == 1 then
				level.setTile(newX, newY, level.getTile(x, y))
				level.setTile(x, y, tiles.BLANK)
				return nil
			else
				return hitTile
			end
		end
	end,
	moveMirror = function(x, y, dx, dy)
		local newMirror = level.moveTile(x, y, dx, dy)
		if newMirror then
			local newX = x + dx
			local newY = y + dy
			local oldMirror = level.getTile(x, y)
			if (oldMirror == tiles.MIRROR_NW and newMirror == tiles.MIRROR_SE) or (oldMirror == tiles.MIRROR_SE and newMirror == tiles.MIRROR_NW) then
				level.setTile(newX, newY, tiles.MIRROR_NW_SE)
				level.setTile(x, y, tiles.BLANK)
				sounds.play("grate")
			elseif (oldMirror == tiles.MIRROR_NE and newMirror == tiles.MIRROR_SW) or (oldMirror == tiles.MIRROR_SW and newMirror == tiles.MIRROR_NE) then
				level.setTile(newX, newY, tiles.MIRROR_NE_SW)
				level.setTile(x, y, tiles.BLANK)
				sounds.play("grate")
			else
				sounds.play("bzzt")
			end
		else
			sounds.play("grate")
		end
	end,
	moveTrackTank = function(buttonX, buttonY, dx, dy)
		local tankDistance = 0
		local tankFindTile
		local x = buttonX
		local y = buttonY
		while true do
			x = x - dx
			y = y - dy
			tankDistance = tankDistance + 1
			if x < 0 or x >= level.width or y < 0 or y >= level.height then
				return false
			else
				tankFindTile = level.getTile(x, y)
				if tankFindTile >= tiles.TRACK_TANK_N and tankFindTile <= tiles.TRACK_TANK_W then
					if tankDistance > 1 then
						local oldTrack = level.getTile(x + dx, y + dy)
						if oldTrack == tiles.TRACK_NS or oldTrack == tiles.TRACK_WE then
							level.setTile(x + dx, y + dy, tankFindTile)
							level.setTile(x, y, oldTrack)
							level.setTemporaryTile(buttonX, buttonY, level.getTile(buttonX, buttonY) + (tiles.ARROW_N_PRESSED - tiles.ARROW_N))
							sounds.play("clck")
							return true
						else
							return false
						end
					else
						return false
					end
				elseif tankFindTile ~= tiles.TRACK_NS and tankFindTile ~= tiles.TRACK_WE and tankFindTile ~= tiles.TRACK_NS_DAMAGED and tankFindTile ~= tiles.TRACK_WE_DAMAGED then
					return false
				end
			end
		end
		return false
	end,
	draw = function()
		for y = 0, level.height - 1 do
			for x = 0, level.width - 1 do
				tile = string.byte(level.rows[y + 1], x + 1)
				tiles.images[tile + 1]:draw(x * tiles.size, y * tiles.size)
			end
		end
	end,
	click = function(x, y)
		local tile = tiles.definitions[level.getTile(x, y)]
		if tile.click then
			tile.click(x, y)
		end
	end
}